package Go;

import Go.*;
import java.util.Vector;
import java.util.Enumeration;
import roulette.*;

// This class models a "greedy" Atari Go player:
// captures if he can, defend against direct capture (if necessary)
// otherwise just plays according to a given heuristic
public class AtariGoGreedyPlayer extends Player {

  public AtariGoGreedyPlayer(int color) {
    super(color);
  }

  // override Player's initPlayer method
  public void initPlayer() {}

  // Methods declared in GoGameListener interface,
  // implemented by Player
  public void positionChanged(GoGameEvent goGameEvent) {}
  public void gameOver(GoGameEvent goGameEvent) {}

  // override Player's move method
  protected Move move(GoPosition goPosition) {
    Move theMove = null;

    if ((theMove = tryCapture(goPosition)) != null)
      return theMove;
    else if ((theMove = tryDefend(goPosition)) != null)
      return theMove;
    else if ((theMove = attackWeakestGroup(goPosition)) != null)
      return theMove;
    else
      return makeRandomMove(goPosition);
  }

  // try to capture an enemy group
  protected Move tryCapture(GoPosition goPosition) {
    ListOfGroups listOfNeighborEnemies;
    ListOfLiberties libertiesList;

    if (this.color == GoConstants.BLACK)
      listOfNeighborEnemies = goPosition.getWhiteGroups();
    else
      listOfNeighborEnemies = goPosition.getBlackGroups();

    Enumeration enemies = listOfNeighborEnemies.getElements();
    for (;enemies.hasMoreElements();) {
      libertiesList = ((Group)enemies.nextElement()).liberties;
      if (libertiesList.length() == 1) {
        GobanLocation location = libertiesList.getElementAt(0);
        return new Move(location.x, location.y, this.color);
      }
    }
    return null;
  }

  // try to defend own's groups against capture
  protected Move tryDefend(GoPosition goPosition) {
    ListOfGroups listOfNeighborFriends = new ListOfGroups();
    ListOfLiberties libertiesList;
    Move theMove;

    if (this.color == GoConstants.WHITE)
      listOfNeighborFriends = goPosition.getWhiteGroups();
    else
      listOfNeighborFriends = goPosition.getBlackGroups();

    Enumeration friends = listOfNeighborFriends.getElements();
    for (;friends.hasMoreElements();) {
      libertiesList = ((Group)friends.nextElement()).liberties;
      if (libertiesList.length() == 1) {
        GobanLocation location = libertiesList.getElementAt(0);
        theMove = new Move(location.x, location.y, this.color);

        if (goPosition.isLegalMove(theMove))
          return theMove;
      }
    }
    return null;
  }

  // find weakest group overall and attack it!
  protected Move attackWeakestGroup(GoPosition goPosition) {
    Move theMove, bestMove = null;

    ListOfGroups listOfEnemies;
    ListOfGroups listOfFriends;

    ListOfLiberties libertiesList;
    Group currentWeakest = null, currentGroup = null;
    int currentNrLiberties = 1000, bestNrliberties = -1;

    if (this.color == GoConstants.BLACK) {
      listOfEnemies = goPosition.getWhiteGroups();
      listOfFriends = goPosition.getBlackGroups();
    }
    else {
      listOfEnemies = goPosition.getBlackGroups();
      listOfFriends = goPosition.getWhiteGroups();
    }

    Enumeration enemies = listOfEnemies.getElements();
    Enumeration friends = listOfFriends.getElements();

    // find weakest enemy group
    for (;enemies.hasMoreElements();) {
      currentGroup = (Group)enemies.nextElement();
      libertiesList = currentGroup.liberties;

      if (libertiesList.length() < currentNrLiberties) {
        currentNrLiberties = libertiesList.length();
        currentWeakest = currentGroup;
      }
    }

    // find weakest own group
    for (;friends.hasMoreElements();) {
      currentGroup = (Group)friends.nextElement();
      libertiesList = currentGroup.liberties;

      if (libertiesList.length() < currentNrLiberties) {
        currentNrLiberties = libertiesList.length();
        currentWeakest = currentGroup;
      }
    }

    if (currentWeakest == null) return null;

    libertiesList = currentWeakest.liberties;
    for (int i = 0; i < libertiesList.length(); i++) {
      GobanLocation location = libertiesList.getElementAt(i);
        theMove = new Move(location.x, location.y, this.color);

        // consider only the legal moves
        if (goPosition.isLegalMove(theMove)) {
          if ((currentNrLiberties = resultedNrLiberties(goPosition, theMove)) >
               bestNrliberties) {
            // found a better move...remember it!
            bestMove = new Move(theMove.x, theMove.y, this.color);
            bestNrliberties = currentNrLiberties;
          }
        }
    }
    return bestMove;
  }

  private Move makeRandomMove(GoPosition goPosition) {
    Vector possibleMoveList = new Vector();
    Vector possibleMoveValueList = new Vector();

    Move currentMove = null;
    float currentMoveValue;
    int gobanSize = goPosition.getSize();

    // Build the list of possible next moves and their estimsted values
    for (int i = 0; i < gobanSize; i++)
      for (int k = 0; k < gobanSize; k++) {
        if (goPosition.goban[i][k].state == GoConstants.EMPTY) {
          currentMove = new Move(this.color);
          currentMove.x = i;
          currentMove.y = k;

          if (goPosition.isLegalMove(currentMove)) {

            goPosition.goban[i][k].state = color;

            currentMoveValue = constantMoveValue;

            possibleMoveList.addElement(currentMove);

            // every legal move is given an equal chance to be choosen
            possibleMoveValueList.addElement(new Double(currentMoveValue));

            // restore original position
            goPosition.goban[i][k].state = GoConstants.EMPTY;
          }
        }
      }

    if (possibleMoveList.size() > 0) {
      // Choose a move in the list probabilistically
      int choosenMoveIndex = Roulette.rouletteChooseOne(possibleMoveValueList);

      return (Move)possibleMoveList.elementAt(choosenMoveIndex);
    }
    // if there is no possible move, play Pass
    else
      return new Move(GoConstants.PASS, GoConstants.PASS, this.color);
  }

  // counts the number of liberties lastMove would have if played in goPosition
  private int resultedNrLiberties(GoPosition goPosition, Move lastMove) {
    int x, y, oldColor;
    int newOwnGroupNrLiberties;

    x = lastMove.x;
    y = lastMove.y;

    oldColor = goPosition.goban[x][y].state;
    goPosition.goban[x][y].state = lastMove.color;
    try {
      GobanLocation lastMoveLocation = goPosition.goban[x][y];

      ListOfGroups listOfNeighborFriends = new ListOfGroups();
      ListOfGroups listOfNeighborEnemies = new ListOfGroups();
      // fill the lists with new stone's friend and enemy groups
      goPosition.visitNewStoneNeighbors(lastMoveLocation,
                                        listOfNeighborFriends,
                                        listOfNeighborEnemies);

      newOwnGroupNrLiberties = game.goPosition.countLiberties(x, y);

      Enumeration friends = listOfNeighborFriends.getElements();
      for (;friends.hasMoreElements();) {
        newOwnGroupNrLiberties += ((Group)friends.nextElement()).liberties.length();
        newOwnGroupNrLiberties -= 1;
      }

    } finally {goPosition.goban[x][y].state = oldColor;};

    return newOwnGroupNrLiberties;
  }

  private final float constantMoveValue = 1;
}
